#include "stdafx.h"
#include "TFMaker.hpp"
#include "FileTypeDeterminer.hpp"
#include "IRProviderTSP.hpp"
#include "IRProviderIR.hpp"
#include "utils.hpp"
#include "SrcMicPair.hpp"
#include "OutputType.hpp"
#include "OutputFormat.hpp"
#include "Incrementer.hpp"
#include "CommonVariables.hpp"
#include <stdexcept>
#include <algorithm>
#include <complex>
#include <chrono>
#include <cmath>
#include <boost/filesystem.hpp>
#include <boost/optional.hpp>
#include <boost/format.hpp>
#include <boost/make_shared.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <harkio2/harkio.hpp>
#include "fft.hpp"

#include "MultiDimAlgorithm.hpp"
#include "FunctionObjects.hpp"

#include <iostream>
#include <string>

#include "debug.hpp"


namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace harkc = harkio2::harkconfig;
namespace hark = harkio2::hark;

namespace
{

// 音源構造体へのポインタの型
typedef boost::shared_ptr<harkc::Source> SourcePtr;
void ThrowSourceVersionException(unsigned int version)
{
    throw std::runtime_error(
            (boost::format("unknown Source structure version: %d")
             % version).str());
}

// 作成する伝達関数の種類
enum class TFType
{
    M2PG, GTF
};

struct TFMakingParameters
{
    std::string ir_list_file;
    std::string tsp_list_file;
    std::string mic_location_file;
    TFType tf_type;
    unsigned int nfft;
    unsigned int tsp_length;
    bool normalize_src;
    bool normalize_mic;
    bool normalize_freq;
    unsigned int direct_sound_length;
    unsigned int sampling_freq;
    unsigned int num_height;
	unsigned int calc;       //geometry-calculus
    

    TFMakingParameters()
        : ir_list_file(), tsp_list_file(), mic_location_file(),
          tf_type(TFType::M2PG), nfft(0), tsp_length(0),
          normalize_src(false), normalize_mic(false), normalize_freq(false),
          direct_sound_length(0)
    {}
};




// (0, 0, -1)方向の音源かどうか
bool IsSpecialDirection(SourcePtr source)
{
    auto version = source->GetVersion();
    if (version == 2)
    {
        auto source_ = boost::dynamic_pointer_cast<harkc::Source2>(source);
        const auto& v = source_->GetPosition().GetPositionVector();
        return v.GetX() == 0 && v.GetY() == 0 && v.GetZ() == -1;
    }
    ThrowSourceVersionException(version);
}

template <typename T>
struct Abs
{
    T operator ()(const T& value) const
    {
        return std::abs(value);
    }
};

template <typename T>
struct Conj
{
    T operator ()(const T& value) const
    {
        return std::conj(value);
    }
};

template <typename T>
class NPower
{
    T power_;
public:
    NPower(const T& power)
        : power_(power)
    {}

    template <typename U>
    auto operator ()(const U& value) const -> decltype(std::pow(value, T()))
    {
        return std::pow(value, power_);
    }
};

/** @brief dstが空ならsrcをdstにコピーし、trueを返す.
 * dst.empty()がtrueの場合、srcをdstにコピーし、trueを返す。
 * そうでなければ、コピーしないでfalseを返す。
 */
template <typename T>
bool IfEmptyThenSet(T& dst, const T& src)
{
    if (dst.empty())
    {
        dst = src;
        return true;
    }
    else
    {
        return false;
    }
}

/** @brief インパルス応答の集合を表すresultへ、ir_providerからインパルス応答を読み込む.
 * resultは nsrc x nmic x nfft でなければならない。
 */
harktool4::MultiDimVector<std::vector, float, 3>::type
LoadIR(harkc::MicArrayLocationFile2_0& mic_location_file,
        harktool4::IRProvider& ir_provider,
        const std::vector<SourcePtr>& sources,
        unsigned int nfft)
{
    //std::map<SrcMicPair, std::vector<float> > result;
    auto nmic = ir_provider.GetNumMic();
    auto nsrc = ir_provider.GetNumSrc();

    // nsrc x nmic x nfft の多次元配列を生成
    auto result = harktool4::CreateMultiDimVector<std::vector, float>(0.0f, nsrc, nmic, nfft);

    for (size_t src = 0; src < nsrc; ++src)
    {
        unsigned int fft_offset;
        auto version = sources[src]->GetVersion();
        if (version == 2)
        {
            fft_offset = boost::dynamic_pointer_cast<harkc::Source2>(sources[src])->GetCutStart() - 1;
        }
        else
        {
            ThrowSourceVersionException(version);
        }
        for (size_t mic = 0; mic < nmic; ++mic)
        {
            auto& ir_vector = result[src][mic];
            assert(ir_vector.size() == nfft);

            // src, micに対応するインパルス応答を取得
            const auto& ir = ir_provider.GetIR(src, mic);
            for (size_t fft = 0; fft < nfft; ++fft)
            {
                ir_vector[fft] = ir.at(fft + fft_offset);
            }
        }
    }

    return result;
}

/** @brief 音源ごとにインパルス応答を正規化する.
 */
void Normalize(harktool4::MultiDimVector<std::vector, float, 3>::type& ir_vect)
{
    using namespace harktool4;

    // ir_vectを一時変数へコピー
    std::remove_reference<decltype(ir_vect)>::type temp_ir_vect(ir_vect);

    MultiDimForEach(temp_ir_vect, MakeInplaceFunctor(NPower<float>(2)),
            md_all,
            md_all,
            md_all);

    for (size_t src = 0; src < ir_vect.size(); ++src)
    {
        const auto sum = harktool4::SumAll(temp_ir_vect[src]);
        harktool4::DivideAll(ir_vect[src], sum);
    }
    std::cout << "Normalize: " << ir_vect[0][0][0] << ", " << ir_vect[1][2][3] << std::endl;
}

/** @brief 伝達関数を正規化する.
 */
void Normalize(harktool4::MultiDimVector<std::vector, std::complex<float>, 3>::type& tf_vect,
        unsigned int nfft, bool normalize_src, bool normalize_mic, bool normalize_freq)
{
    using namespace harktool4;

    assert(tf_vect.size() == 0 || tf_vect[0].size() == 0 || tf_vect[0][0].size() == 0 || tf_vect[0][0].size() == nfft / 2 + 1);

    // 各TFの最後の要素を0にする
    MultiDimFill(tf_vect, 0,
            md_all,
            md_all,
            MultiDimRange(nfft/2, nfft/2 + 1));

    if (!normalize_src && !normalize_mic && !normalize_freq)
    {
        return;
    }

    // tf_vectを一時変数へコピー
    //std::remove_reference<decltype(tf_vect)>::type temp_tf_vect(tf_vect);
    std::cout << "tf_vect size: " << tf_vect.size() << ", " << tf_vect[0].size() << ", " << tf_vect[0][0].size() << std::endl;
    harktool4::MultiDimVector<std::vector, std::complex<float>, 3>::type temp_tf_vect(tf_vect);

    MultiDimForEach(temp_tf_vect, MakeInplaceFunctor(Abs<std::complex<float>>()),
            md_all,
            md_all,
            md_all);
    MultiDimForEach(temp_tf_vect, MakeInplaceFunctor(NPower<float>(2)),
            md_all,
            md_all,
            md_all);

    for (size_t src = 0; src < (normalize_src ? 1 : tf_vect.size()); ++src)
    {
        for (size_t mic = 0; mic < (normalize_mic ? 1 : tf_vect[src].size()); ++mic)
        {
            for (size_t freq = 0; freq < (normalize_freq ? 1 : tf_vect[src][mic].size()); ++freq)
            {
                auto src_range = MultiDimRange(src, normalize_src ? tf_vect.size() : (src + 1));
                auto mic_range = MultiDimRange(mic, normalize_mic ? tf_vect[src].size() : (mic + 1));
                auto freq_range = MultiDimRange(freq, normalize_freq ? tf_vect[src][mic].size() : (freq + 1));

                auto sum = MultiDimAccumulate(temp_tf_vect, std::complex<float>(0), std::plus<std::complex<float>>(),
                        src_range, mic_range, freq_range).real();
                auto sqrt_sum = sqrt(sum);
                if (std::isfinite(sqrt_sum) && sqrt_sum != 0)
                {
                    MultiDimForEach(tf_vect, MakeInplaceFunctor(std::bind2nd(std::divides<std::complex<float>>(), sqrt_sum)),
                            src_range, mic_range, freq_range);
                }
            }
        }
    }
}

/** @brief (0, 0, -1)の音源を削除
 */
void Erase001(harktool4::MultiDimVector<std::vector, float, 3>::type& ir_vect, std::vector<SourcePtr>& sources)
{
    auto srcit = sources.begin();
    auto irit = ir_vect.begin(); // sourcesの最上位の次元は音源番号なので、最上位の次元で繰り返す
    while (srcit != sources.end())
    {
        if (IsSpecialDirection(*srcit))
        {
            // 音源方向が(0, 0, -1)ならその音源情報と、対応するインパルス応答を削除
            srcit = sources.erase(srcit);
            irit = ir_vect.erase(irit);
        }
        else
        {
            ++srcit;
            ++irit;
        }
    }
}

/** @brief 残響音を取り除く.
 * @param ir_vect NSRC x NMIC x NFFTの3次元配列
 * @param sources 音源配列
 * @param direct_sound_length 直接音とみなす信号長
 */
void EraseReverberation(
        harktool4::MultiDimVector<std::vector, float, 3>::type& ir_vect,
        const std::vector<SourcePtr>& sources,
        unsigned int direct_sound_length)
{
    for (size_t src = 0; src < ir_vect.size(); ++src)
    {
        unsigned int cut_end;
        auto version = sources[src]->GetVersion();
        if (version == 2)
        {
            cut_end = boost::dynamic_pointer_cast<harkc::Source2>(sources[src])->GetCutEnd();
        }
        else
        {
            ThrowSourceVersionException(version);
        }

        for (size_t mic = 0; mic < ir_vect[src].size(); ++mic)
        {
            auto& ir_vector = ir_vect[src][mic];
            auto fill_start_pos = ir_vector.begin();

            if (0 < cut_end)
            {
                if (ir_vector.size() <= cut_end)
                {
                    throw std::runtime_error(
                            (boost::format("cut_end exceeds the number of FFT: src %d") % src).str());
                }
                std::advance(fill_start_pos, cut_end);
            }
            else
            {
                // ir_vectorの絶対値を求め、abs_irに格納
                std::vector<float> abs_ir(ir_vector.size());
                std::transform(ir_vector.begin(), ir_vector.end(), abs_ir.begin(), static_cast<float(*)(float)>(std::abs));
                // 最大値の位置を探す
                auto top_iterator = harktool4::FindTop(abs_ir.begin(), abs_ir.end());
                auto top_index = std::distance(abs_ir.begin(), top_iterator);

                // 反響音の位置distを計算し、distの範囲をチェック
                auto dist = top_index + 1 + direct_sound_length;
                if (dist < ir_vector.size())
                {
                    std::advance(fill_start_pos, dist);
                }
                else
                {
                    fill_start_pos = ir_vector.end();
                }
            }

            // ir_vectorの反響音をクリア
            std::fill(fill_start_pos, ir_vector.end(), 0);
        }
    }
}

std::vector<harktool4::SrcMicPair>
GetTooStrongReverbSources(
        harktool4::MultiDimVector<std::vector, float, 3>::type& ir_vect,
        unsigned int direct_sound_position_limit)
{
    std::vector<harktool4::SrcMicPair> result;

    for (size_t src = 0; src < ir_vect.size(); ++src)
    {
        for (size_t mic = 0; mic < ir_vect[src].size(); ++mic)
        {
            auto& ir_vector = ir_vect[src][mic];

            // ir_vectorの絶対値を求め、abs_irに格納
            std::vector<float> abs_ir(ir_vector.size());
            std::transform(ir_vector.begin(), ir_vector.end(), abs_ir.begin(), static_cast<float(*)(float)>(std::abs));
            // 最大値の位置を探す
            auto top_iterator = harktool4::FindTop(abs_ir.begin(), abs_ir.end());
            auto top_index = std::distance(abs_ir.begin(), top_iterator);

            // インパルス応答の最大値（直接音）が大きすぎる場合を記録
            if (direct_sound_position_limit <= static_cast<unsigned int>(top_index))
            {
                result.emplace_back(src, mic);
            }
        }
    }

    return result;
}

void ShowTooStrongReverbSources(
        const std::vector<harktool4::SrcMicPair>& reverb_source_indices,
        const std::vector<SourcePtr>& sources)
{
    for (auto it = reverb_source_indices.begin(); it != reverb_source_indices.end(); ++it)
    {
        //const auto& source_ = sources.at(it->src_index);
        auto version = sources.at(it->src_index)->GetVersion();

        if (version == 2)
        {
            auto source = boost::dynamic_pointer_cast<harkc::Source2>(sources.at(it->src_index));
            const auto& files = source->GetFiles();
            for (auto fileit = files.begin(); fileit != files.end(); ++fileit)
            {
                const auto& mic_indices = fileit->GetMicIndices();
                for (auto micit = mic_indices.begin(); micit != mic_indices.end(); ++micit)
                {
                    if (*micit == it->mic_index)
                    {
                        std::cout << fileit->GetName()
                            << " src " << it->src_index << ", mic " << it->mic_index << '\n';
                    }
                }
            }
        }
        else
        {
            ThrowSourceVersionException(version);
        }
    }
}

harktool4::MultiDimVector<std::vector, std::complex<float>, 3>::type
DoFFT(const harktool4::MultiDimVector<std::vector, float, 3>::type& ir_vect, unsigned int nfft)
{
    auto result_vect = harktool4::CreateMultiDimVector<std::vector, std::complex<float>>(
            std::complex<float>(),
            ir_vect.size(), ir_vect.front().size(), nfft / 2 + 1);
    HRI::fft<float> fft_obj(nfft);
    for (unsigned int src = 0; src < ir_vect.size(); src++) {
        for (unsigned int mic = 0; mic < ir_vect[src].size(); mic++) {
            assert(ir_vect[src][mic].size() == nfft);

            std::copy_n(ir_vect[src][mic].begin(), nfft, fft_obj.a.begin());
            //for (unsigned int fft = 0; fft < nfft; fft++) {
            //    fft_obj.a[fft] = matrix.matrix[fft][mic][src];
            //}
            fft_obj.analyze(); // fft実行
            std::copy_n(fft_obj.A.begin(), nfft / 2 + 1, result_vect[src][mic].begin());
            //for (unsigned int fft = 0; fft < nfft; fft++) {
            //    matrix_H.matrix[fft][mic][src] = fft_obj.A[fft];
            //}
        }
    }

    return result_vect;
}

void Conjugate(harktool4::MultiDimVector<std::vector, std::complex<float>, 3>::type& tf_vect)
{
    using namespace harktool4;

    MultiDimForEach(tf_vect, MakeInplaceFunctor(Conj<std::complex<float>>()),
            md_all,
            md_all,
            md_all);
}

std::vector<harkio2::Vector3<float>>
GetMicPositions(const std::vector<harkc::Mic1>& mics)
{
    std::vector<harkio2::Vector3<float>> result(mics.size());
    for (size_t i = 0; i < result.size(); ++i)
    {
        const auto& mic = mics[i];
        result[mic.GetMicIndex()] = mic.GetPosition().GetPositionVector();
    }
    return result;
}

std::vector<harkio2::Vector3<float>>
GetSourcePositions(const std::vector<SourcePtr>& sources)
{
    std::vector<harkio2::Vector3<float>> result(sources.size());
    for (size_t i = 0; i < result.size(); ++i)
    {
        auto version = sources[i]->GetVersion();
        if (version == 2)
        {
            const auto& src = boost::dynamic_pointer_cast<harkc::Source2>(sources[i]);
            result[i] = src->GetPosition().GetPositionVector();
        }
        else
        {
            ThrowSourceVersionException(version);
        }
    }
    return result;
}

harktool4::MultiDimVector<std::vector, hark::hgtf::M2PG1_0::TFType, 3>::type
ConvertM2PG(const harktool4::MultiDimVector<std::vector, std::complex<float>, 3>::type& tf_vect, size_t nd, size_t nr)
{
    const auto nmic = tf_vect.front().size();
    auto result = harktool4::CreateMultiDimVector<std::vector, hark::hgtf::M2PG1_0::TFType>(
            hark::hgtf::M2PG1_0::TFType(tf_vect.front().front().size()),
            nd, nr, nmic);
    for (size_t d = 0; d < nd; ++d)
    {
        for (size_t r = 0; r < nr; ++r)
        {
            for (size_t mic = 0; mic < nmic; ++mic)
            {
                const auto& temp = tf_vect[d * nr + r][mic];
                std::copy(temp.begin(), temp.end(), result[d][r][mic].begin());
            }
        }
    }

    return result;
}

harktool4::MultiDimVector<std::vector, harkio2::Vector3<float>, 2>::type
ConvertM2PG(const std::vector<harkio2::Vector3<float>>& sources, size_t nd, size_t nr)
{
    auto result = harktool4::CreateMultiDimVector<std::vector, harkio2::Vector3<float>>(
            harkio2::Vector3<float>(),
            nd, nr);
    for (size_t d = 0; d < nd; ++d)
    {
        for (size_t r = 0; r < nr; ++r)
        {
            result[d][r] = sources[d * nr + r];
        }
    }

    return result;
}




//-----------------------------------------------------------------------//
//
//   ConvertToHarkFile
//
//-----------------------------------------------------------------------//
boost::shared_ptr<hark::HarkFile>
ConvertToHarkFile(
        const harktool4::MultiDimVector<std::vector, std::complex<float>, 3>::type& tf_vect,
        const std::vector<harkio2::Vector3<float>>& sources,
        boost::shared_ptr<harkc::MicArrayLocationFile2_0> mic_location_file,
        unsigned int nfft, // Number of FFT
        unsigned int nir,  // Number of Impulse Response
        unsigned int fs,   // Sampling Frequency
        TFType tf_type)
{

//	DEBUG_PRINT("@ ConvertToHarkFile start: %s[%d]\n", __FILE__ ,__LINE__);


    using namespace harkio2::hark;
    auto result = boost::make_shared<HarkFile>();
    auto fmt = boost::make_shared<hgtf::Fmt1_0>();
    auto spec = boost::make_shared<hgtf::Spec1_0>();
    boost::shared_ptr<DataSet> m2pg_or_gtf;

    if (tf_vect.empty() || tf_vect.front().empty() || tf_vect.front().front().empty())
    {
        throw std::runtime_error("Empty TF vector");
    }

    auto mics = mic_location_file->GetMics();
    auto now_time = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
    auto now_local_time = localtime(&now_time);

    fmt->SetFormatId(tf_type == TFType::M2PG ? 1 : 2); // M2PG or GTF
    fmt->SetNMic(mics.size());
    fmt->SetNFFT(nfft);
    fmt->SetNImpulseResponse(nir);
    fmt->SetFS(fs);
    fmt->SetMicPositions(GetMicPositions(mic_location_file->GetMics()));

    spec->SetDate((boost::format("%04d/%02d/%02d")
                % (now_local_time->tm_year + 1900) % (now_local_time->tm_mon + 1) % now_local_time->tm_mday).str());
    spec->SetLocation("");
    spec->SetRobot("");
    spec->SetMicArrayType("");
    {
        const char* p;
        ;
        spec->SetCreator(
                (p = getenv("USERNAME")) ? p
                : (p = getenv("USER")) ? p
                : "");
    }
    spec->SetMemo("");

    if (tf_type == TFType::M2PG)
    {
        auto m2pg = boost::make_shared<hgtf::M2PG1_0>();
        m2pg->SetND(tf_vect.size());
        m2pg->SetNR(1);
        m2pg->SetSourcePositions(ConvertM2PG(sources, sources.size(), 1));
        m2pg->SetTFs(ConvertM2PG(tf_vect, sources.size(), 1));

        m2pg_or_gtf = m2pg;
    }
    else
    {
        auto gtf = boost::make_shared<hgtf::GTF1_0>();
        std::vector<unsigned int> findex_to_binindex(nfft/2 + 1);
        // findex_to_binindex := 0, 1, 2, ..., nfft/2+1
        //std::iota(findex_to_binindex.begin(), findex_to_binindex.end(), 0);
        for (unsigned int i = 0; i < findex_to_binindex.size(); ++i)
        {
            findex_to_binindex[i] = i;
        }

        gtf->SetNSource(tf_vect.size());
        gtf->SetSourcePositions(sources);
        gtf->SetNFIndex(findex_to_binindex.size());
        gtf->SetFIndexToBinIndexTable(findex_to_binindex);
        gtf->SetTFs(tf_vect);

        m2pg_or_gtf = gtf;
    }

    result->AddDataSet(fmt);
    result->AddDataSet(m2pg_or_gtf);
    result->AddDataSet(spec);

    return result;
}




//************************************************************************//
//
//
//************************************************************************//
boost::shared_ptr<hark::HarkFile>
ConvertToHarkFile(
        const harktool4::MultiDimVector<std::vector, std::complex<float>, 3>::type& tf_vect,
        const std::vector<SourcePtr>& sources,
        boost::shared_ptr<harkc::MicArrayLocationFile2_0> mic_location_file,
        unsigned int nfft, // Number of FFT
        unsigned int nir,  // Number of Impulse Response
        unsigned int fs,   // Sampling Frequency
        TFType tf_type)
{
    return ConvertToHarkFile(tf_vect, GetSourcePositions(sources), mic_location_file,
            nfft, nir, fs, tf_type);
}





//************************************************************************//
//
//
//************************************************************************//
boost::shared_ptr<hark::HarkFile>
Make(const TFMakingParameters& param, harktool4::IRProvider& ir_provider, unsigned int nir)
//Make(const std::string& mic_location_file_path,
//        harktool4::IRProvider& ir_provider,
//        const std::vector<SourcePtr>& sources,
//        unsigned int nfft,
//        unsigned int nir,
//        TFType tf_type,
//        bool normalize_src,
//        bool normalize_mic,
//        bool normalize_freq,
//        unsigned int direct_sound_length)
{


	DEBUG_PRINT("@ Make start : %s[%d]\n", __FILE__ ,__LINE__);


    const unsigned int DIRECT_SOUND_POSITION_LIMIT = 200;

    // MicLocationFileを読み込む
    harkc::HarkConfigFileXMLReader mic_location_file_reader(param.mic_location_file);
    auto harkconfig_file = mic_location_file_reader.Read();
    auto mic_location_file =
        boost::dynamic_pointer_cast<harkc::MicArrayLocationFile2_0>(harkconfig_file);

    // インパルス応答をすべて読み込む
    auto sources = ir_provider.GetSources();
    auto ir_vect = LoadIR(*mic_location_file, ir_provider, sources, param.nfft);
    auto new_sources = sources;

    if (param.tf_type == TFType::GTF) // GSS
    {
        Normalize(ir_vect);
    }

    if (param.tf_type == TFType::M2PG) // MUSIC
    {
        // 座標が(0,0,-1)の音源を削除
        Erase001(ir_vect, new_sources);

        // 残響音と思われる部分に0を書き込む
        EraseReverberation(ir_vect, new_sources, param.direct_sound_length);

        // 残響音が直接音より大きい可能性がある音源を表示
        auto reverb_source_indices = GetTooStrongReverbSources(ir_vect, DIRECT_SOUND_POSITION_LIMIT);
        ShowTooStrongReverbSources(reverb_source_indices, new_sources);
    }

    // インパルス応答をFFTして伝達関数を取得
    auto tf_vect = DoFFT(ir_vect, param.nfft);
    // Conjugate(tf_vect);

    if (param.tf_type == TFType::GTF) // GSS
    {
        Conjugate(tf_vect);
        Normalize(tf_vect, param.nfft, param.normalize_src, param.normalize_mic, param.normalize_freq);
    }

    // HARKファイル形式に変換
    return ConvertToHarkFile(tf_vect, new_sources, mic_location_file, param.nfft, nir, param.sampling_freq, param.tf_type);
    //return std::shared_ptr<hark::HarkFile>();
}




//************************************************************************//
//
// 幾何計算から伝達関数を作る
//
//************************************************************************//
boost::shared_ptr<hark::HarkFile>
Make(const TFMakingParameters& param, harktool4::CommonVariables vars)
{

//	DEBUG_PRINT("@ 幾何計算から伝達関数を作る start : %s[%d]\n", __FILE__ ,__LINE__);

    namespace ublas = boost::numeric::ublas;
    using std::vector;

    typedef std::complex<float> fcomplex;
    typedef ublas::vector<float> fvector;
    typedef ublas::matrix<float> fmatrix;
    typedef ublas::vector<fcomplex> fcvector;
    typedef ublas::matrix<fcomplex> fcmatrix;

    if (vars.num_vertical != 1)
    {
        throw std::runtime_error("Geometric TF calculation doesn't support multi vertical sound positions");
    }


    const float sound_speed = 343;


	//---------------------------------------------------//
	// ListFileを読み込む
	//---------------------------------------------------//
    harkc::HarkConfigFileXMLReader tsp_list_file_reader(param.tsp_list_file);
    auto tsp_harkconfig_file = tsp_list_file_reader.Read();
    auto tsp_list_file =
        boost::dynamic_pointer_cast<harkc::TSPListFile2_0>(tsp_harkconfig_file);


	const auto tsp_nummic =  tsp_list_file->GetNumMic();
	const auto& tsp_sources = tsp_list_file->GetSources(); 
	const auto  tsp_numsources = tsp_list_file->GetSources().size(); 







	//---------------------------------------------------//
    // MicLocationFileを読み込む
	//---------------------------------------------------//
    harkc::HarkConfigFileXMLReader mic_location_file_reader(param.mic_location_file);
    auto harkconfig_file = mic_location_file_reader.Read();
    auto mic_location_file =
        boost::dynamic_pointer_cast<harkc::MicArrayLocationFile2_0>(harkconfig_file);

    const auto& mics = mic_location_file->GetMics();
    const auto nmic = mics.size();
    fmatrix mic_positions(nmic, 3);
    for (unsigned int i = 0; i < nmic; ++i)
    {
        const auto& pos = mics[i].GetPosition().GetPositionVector();
        mic_positions(i, 0) = pos.GetX();
        mic_positions(i, 1) = pos.GetY();
        mic_positions(i, 2) = pos.GetZ();
    }

    fvector position_mean(ublas::zero_vector<float>(3));
    for (unsigned int i = 0; i < nmic; ++i)
    {
        position_mean += row(mic_positions, i);
    }
    position_mean /= nmic;
    for (unsigned int i = 0; i < nmic; ++i)
    {
        row(mic_positions, i) -= position_mean;
    }





    vector<vector<vector<fcvector>>> A(
            tsp_numsources , vector<vector<fcvector>>(  // number of directions
                1, vector<fcvector>(                // number of radius
                    nmic, fcvector(                 // number of microphones
                        param.nfft / 2 + 1))));     // number of FFT


    vector<vector<harkio2::Position<float>>> src_positions(
            tsp_numsources , vector<harkio2::Position<float>>(      // number of directions
                1));                                            // number of radius


    vars.azimuth.Reset();





#if 0 //---takenaka
    vector<vector<vector<fcvector>>> A(
            vars.numazi, vector<vector<fcvector>>(  // number of directions
                1, vector<fcvector>(                // number of radius
                    nmic, fcvector(                 // number of microphones
                        param.nfft / 2 + 1))));     // number of FFT

    vector<vector<harkio2::Position<float>>> src_positions(
            vars.numazi, vector<harkio2::Position<float>>(      // number of directions
                1));                                            // number of radius

    vars.azimuth.Reset();
    
   
    //-------------------------------------------------------------------
	// ListFile１を読み込む
	//------------------------------------------------------------------
    harkc::HarkConfigFileXMLReader tsp_list_file_reader(param.tsp_list_file);
    auto tsp_harkconfig_file = tsp_list_file_reader.Read();
    auto tsp_list_file =
        boost::dynamic_pointer_cast<harkc::TSPListFile2_0>(tsp_harkconfig_file);


	const auto tsp_nummic =  tsp_list_file->GetNumMic();
	const auto& tsp_sources = tsp_list_file->GetSources(); 
	const auto  tsp_numsources = tsp_list_file->GetSources().size(); 

#endif

	

	//---------------------------------------------------//
	// ListFileからX、Y、Zを読み込む
	//---------------------------------------------------//
	fmatrix tsp_positions(tsp_numsources , 3);

	const auto& pos = tsp_sources[0].GetPosition().GetPositionVector();


    for (unsigned int dir = 0; dir < tsp_numsources ; ++dir)
    {
		const auto& tsp_pos = tsp_sources[dir].GetPosition().GetPositionVector();

//		DEBUG_PRINT("tsp_pos.GetX()=%f\n",tsp_pos.GetX() );
//	   	DEBUG_PRINT("tsp_pos.GetY()=%f\n",tsp_pos.GetY() );
//	   	DEBUG_PRINT("tsp_pos.GetZ()=%f\n",tsp_pos.GetZ() );

		src_positions[dir][0] = harkio2::MakePositionFromVector<float>( harkio2::Vector3<float>( tsp_pos.GetX(), tsp_pos.GetY(), tsp_pos.GetZ()) );
	}
    


    
    
#if 0 // delete     
    for (unsigned int dir = 0; dir < vars.numazi; ++dir)
    {
        src_positions[dir][0] =
            harktool4::MakePosition(vars.use_height, *vars.azimuth, *vars.height, *vars.elevation, vars.radius);
        ++vars.azimuth;
    }
#endif


    const float sound_speed_1 = param.sampling_freq / sound_speed;


    
    for (unsigned int dir = 0; dir < tsp_numsources; ++dir)
    {
        const auto& pos = src_positions[dir][0].GetPositionVector();
        fvector src_pos_vect(3);
        src_pos_vect[0] = pos.GetX();
        src_pos_vect[1] = pos.GetY();
        src_pos_vect[2] = pos.GetZ();

		//mic
        for (unsigned int mic = 0; mic < nmic; ++mic)
        {
            const auto& mic_pos_vect = row(mic_positions, mic);
            const float src_mic_delta = sound_speed_1 * (- norm_2(src_pos_vect - mic_pos_vect));
			
			//fft
            for (unsigned int fft = 0; fft < param.nfft / 2 + 1; ++fft)
            {
                const float theta = 2 * M_PI * src_mic_delta * fft / param.nfft;
                A[dir][0][mic][fft] = std::complex<float>(cos(theta), sin(theta));
            }
        }

    }




    auto tf_vect = harktool4::CreateMultiDimVector<std::vector, fcomplex>(fcomplex(), tsp_numsources * 1, nmic, param.nfft / 2 + 1);

    for (size_t dir = 0; dir < A.size(); ++dir)
    {
        for (size_t r = 0; r < A[dir].size(); ++r)
        {
            for (size_t mic = 0; mic < A[dir][r].size(); ++mic)
            {
                std::copy(A[dir][r][mic].begin(), A[dir][r][mic].end(), tf_vect[dir * A[dir].size() + r][mic].begin());
            }
        }
    }


    std::vector<harkio2::Vector3<float>> sources(tsp_numsources * 1);
    for (size_t dir = 0; dir < src_positions.size(); ++dir)
    {
        for (size_t r = 0; r < src_positions[dir].size(); ++r)
        {
            sources[dir * src_positions[dir].size() + r] = src_positions[dir][r].GetPositionVector();
        }
    }

    return ConvertToHarkFile(tf_vect, sources, mic_location_file, param.nfft, 0, param.sampling_freq, param.tf_type);



}







std::vector<harkc::Source2>
ConvertToSource2Vector(const std::vector<boost::shared_ptr<harkc::Source>>& sources)
{
    std::vector<harkc::Source2> result(sources.size());

    for (auto it = sources.begin(); it != sources.end(); ++it)
    {
        auto src = boost::dynamic_pointer_cast<harkc::Source2>(*it);
        if (!src)
        {
            throw std::runtime_error(
                    (boost::format("unknown Source structure version: %d")
                     % (*it)->GetVersion()).str());
        }
        result.push_back(*src);
    }

    return result;
}

boost::shared_ptr<harkc::HarkConfigFile>
ReadHarkConfigXMLFile(const std::string& file_path)
{
    harkc::HarkConfigFileXMLReader reader(file_path);
    return reader.Read();
}

} // namespace

namespace harktool4
{

class TFMaker::Impl
{
public:
    Impl(const po::variables_map& vm, std::shared_ptr<std::ostream> output_file)
        : vm_(vm), output_file_(output_file)
    {}

    void Run();

private:
    const po::variables_map& vm_;
    std::shared_ptr<std::ostream> output_file_;

    TFMakingParameters MakeParameters() const;
    //void MakeFromIR(const TFMakingParameters& param) const;
    //void MakeFromTSP(const TFMakingParameters& param) const;
};




//-----------------------------------------------------------------//
//  Run
//-----------------------------------------------------------------//
void TFMaker::Impl::Run()
{


	DEBUG_PRINT("@ TFMaker::Impl::Run start : %s[%d]\n", __FILE__ ,__LINE__);

    OutputFormat output_format = OutputFormat::Binary;
    if (vm_.count("output-format"))
    {
        output_format = vm_["output-format"].as<OutputFormat>();
    }

    std::unique_ptr<hark::HarkFileWriter> writer;
    switch (output_format)
    {
    case OutputFormat::Binary:
        writer.reset(new hark::HarkFileBinaryWriter(*output_file_));
        break;
    default:
        std::cerr << "unsupported output format in TF mode" << std::endl;
        return;
    }

    try
    {

        auto param = MakeParameters();
        CommonVariables vars(vm_);
        std::unique_ptr<IRProvider> ir_provider;
        unsigned int nir = 0;


		if(param.calc == 1 )
		{
			DEBUG_PRINT(" TFMaker::Impl::Run 0 ( 幾何計算): %s[%d]\n", __FILE__ ,__LINE__);
		    writer->Write(Make(param, vars));
		}
		else
		{

		    if (!param.ir_list_file.empty())
		    {
		        ir_provider.reset(new IRProviderIR(param.ir_list_file));
		        nir = ir_provider->GetIR(0, 0).size();
		    }
		    else if (!param.tsp_list_file.empty())
		    {
		        ir_provider.reset(new IRProviderTSP(param.tsp_list_file));
		        nir = param.tsp_length;
		    }

		    if (ir_provider)
		    {
		        writer->Write(Make(param, *ir_provider, nir));
		    }
		    else
		    {
		        writer->Write(Make(param, vars));
		    }
		}

    }
    catch (const std::runtime_error& e)
    {
        std::cerr << e.what() << std::endl;
    }
}




//--------------------------------------------------------------------------------//
/** @brief TF作成用パラメータをvm_から生成し、返す.
 * TF作成が不可能となる程のエラーの場合はstd::runtime_error例外を投げる。
 */
//--------------------------------------------------------------------------------//
TFMakingParameters TFMaker::Impl::MakeParameters() const
{

	DEBUG_PRINT("@ MakeParameters start : %s[%d]\n", __FILE__ ,__LINE__);


    TFMakingParameters param;

	param.calc = vm_["geometry-calculus"].as<unsigned int>() ; //geometry-calculus取得(1:幾何計算時)
	
	
    if (vm_.count("input-files") != 0)
    {
        const std::vector<std::string>& input_files = vm_["input-files"].as<std::vector<std::string>>();

        for (auto it = input_files.begin(); it != input_files.end(); ++it)
        {
            switch (FileTypeDeterminer().Determine(*it))
            {
            case FileType::XML_IR:
                if (IfEmptyThenSet(param.ir_list_file, *it) == false)
                {
                    throw std::runtime_error("too many impulse response list file");
                }
                break;
            case FileType::XML_TSP:
                if (IfEmptyThenSet(param.tsp_list_file, *it) == false)
                {
                    throw std::runtime_error("too many tsp list file");
                }
                break;
            case FileType::XML_MIC:
                if (IfEmptyThenSet(param.mic_location_file, *it) == false)
                {
                    throw std::runtime_error("too many mic array location file");
                }
                break;
            default:
                std::cerr << "the file has a invalid type for making TF. ignoring: " << *it << std::endl;
            }
        }
    }

    param.tf_type = TFType::M2PG;
    
    if (vm_.count("output-type") != 0)
    {
        switch (vm_["output-type"].as<harktool4::OutputType>())
        {
        case harktool4::OutputType::M2PG: param.tf_type = TFType::M2PG; break;
        case harktool4::OutputType::GTF: param.tf_type = TFType::GTF; break;
        default:
            throw std::runtime_error("invalid output type for making TF");
        }
    }

    
    if (param.mic_location_file.empty())
    {
        throw std::runtime_error("need a mic array location file for making TF");
    }
    else if (!param.ir_list_file.empty() && !param.tsp_list_file.empty())
    {
        throw std::runtime_error("too many HARK-config files for making TF");
    }
    else if (param.ir_list_file.empty() && param.tsp_list_file.empty())
    {
        throw std::runtime_error("need a impulse response list file or a tsp list file for making TF");
    }
    else if (param.calc == 1)
    {
        if (param.tf_type == TFType::M2PG)
        {
            std::cerr << "creating TF based on geometric calculation(param.calc == 1)" << std::endl;
        }
        else
        {
            std::cerr << "GTF creating TF based on geometric calculation(param.calc == 1)" << std::endl;
            //throw std::runtime_error("param error : param.calc == 1");
        }
    }    
    
    
    assert((!param.ir_list_file.empty() && param.tsp_list_file.empty()) ||
           (param.ir_list_file.empty() && !param.tsp_list_file.empty()) ||
           (param.ir_list_file.empty() && param.tsp_list_file.empty() && param.tf_type == TFType::M2PG));



    param.nfft = 512;

    if (vm_.count("nfft") != 0)
    {
        param.nfft = vm_["nfft"].as<unsigned int>();
    }

    param.tsp_length = vm_["tsp-length"].as<unsigned int>();

    param.normalize_src = param.normalize_mic = param.normalize_freq = false;
    if (vm_.count("normalize-src") != 0) { param.normalize_src = true; }
    if (vm_.count("normalize-mic") != 0) { param.normalize_mic = true; }
    if (vm_.count("normalize-freq") != 0) { param.normalize_freq = true; }

    param.direct_sound_length = vm_["direct-length"].as<unsigned int>();

    param.sampling_freq = vm_["sampling-freq"].as<unsigned int>();


	DEBUG_PRINT("ir_list_file        = %s \n", param.ir_list_file.c_str());
	DEBUG_PRINT("tsp_list_file       = %s \n", param.tsp_list_file.c_str());
	DEBUG_PRINT("mic_location_file   = %s \n", param.mic_location_file.c_str());
	DEBUG_PRINT("nfft                = %d \n", param.nfft);
	DEBUG_PRINT("tsp_length          = %d \n", param.tsp_length);
	DEBUG_PRINT("normalize_src       = %d \n", param.normalize_src);
	DEBUG_PRINT("normalize_mic       = %d \n", param.normalize_mic );
	DEBUG_PRINT("normalize_freq      = %d \n", param.normalize_freq);
	DEBUG_PRINT("direct_sound_length = %d \n", param.direct_sound_length);
	DEBUG_PRINT("sampling_freq       = %d \n", param.sampling_freq);
//	DEBUG_PRINT("num_height          = %d \n", param.num_height);
	DEBUG_PRINT("calc                = %d \n", param.calc);

    return param;
}





#if 0
void TFMaker::Impl::MakeFromIR(const TFMakingParameters& param) const
{
    IRProviderIR ir_provider(param.ir_list_file);

    auto hark_file = Make(param, ir_provider);

}

void TFMaker::Impl::MakeFromTSP(const TFMakingParameters& param) const
{
    IRProviderTSP ir_provider(param.tsp_list_file);

    auto hark_file = Make(param.mic_location_file, ir_provider, ir_provider.GetSources(), param.nfft, param.tsp_length, param.tf_type,
            param.normalize_src, param.normalize_mic, param.normalize_freq, param.direct_sound_length);

    hark::HarkFileBinaryWriter writer(*output_file_);
    writer.Write(hark_file);
}
#endif

TFMaker::TFMaker(const boost::program_options::variables_map& vm, std::shared_ptr<std::ostream> output_file)
    : impl_(new Impl(vm, output_file))
{}

void TFMaker::Run()
{
    impl_->Run();
}

} // namespace harktool4

